# Standard imports
from typing import Callable
from datetime import datetime


def __now() -> str:
    """ Returns the current datetime as string """
    return datetime.now().strftime('%Y-%m-%d %H:%M:%S')


class _MethodDecoratorAdaptor(object):

    def __init__(self, decorator, func):
        self.decorator = decorator
        self.func = func


    def __call__(self, *args, **kwargs):
        return self.decorator(self.func)(*args, **kwargs)


    def __get__(self, instance, owner):
        return self.decorator(self.func.__get__(instance, owner))



def auto_adapt_to_methods(decorator):
    """ Allows you to use the same decorator on methods and functions, hiding the self argument from the decorator """
    def adapt(func):
        return _MethodDecoratorAdaptor(decorator, func)
    return adapt



def restrict_by_username(whitelist: list = [], unauthorized_reply_msg: str = None):
    """
        A decorator that restricts access to a function based on the username of the user who called it.

        Parameters:
            whitelist (list): A list of usernames that are allowed to access the function.
            unauthorized_reply_msg (str): If not spesified, the default message is replied
    """
    @auto_adapt_to_methods
    def wrapper(func: Callable):
        def wrapped(*args, **kwargs):
            update = args[0]
            username = update.effective_user.username

            if username not in whitelist:
                print(f"[ACCESS DENIED - {__now()}] Unauthorized access denied for '{username}'.")
                update.message.reply_text(unauthorized_reply_msg if unauthorized_reply_msg is not None else f" Sorry, you are not permitted to use this bot.")
                return
            return func(*args, **kwargs)
        return wrapped
    return wrapper



def restrict_by_id(whitelist: list = [], unauthorized_reply_msg: str = None):
    """
        A decorator that restricts access to a function based on the ID of the user who called it.

        Parameters:
            whitelist (list): A list of IDs that are allowed to access the function.
            unauthorized_reply_msg (str): If not spesified, the default message is replied
    """
    @auto_adapt_to_methods
    def wrapper(func: Callable):
        def wrapped(*args, **kwargs):
            update = args[0]
            id = update.effective_user.id

            if id not in whitelist:
                print(f"[ACCESS DENIED - {__now()}] Unauthorized access denied for '{id}'.")
                update.message.reply_text(unauthorized_reply_msg if unauthorized_reply_msg is not None else f" Sorry, you are not permitted to use this bot.")
                return
            return func(*args, **kwargs)
        return wrapped
    return wrapper



# Standard imports
from sqlalchemy import create_engine
from urllib.parse import quote_plus as urlquote

# Third-party imports
from ._sqlalchemy import SQLAlchemy


class PostgresClient(SQLAlchemy):

    def __init__(self, host: str, port: str, username: str, password: str, database: str, schema: str = "public") -> None:
        """
            Creates and initializes a PostgreSQL engine instance that connects to the database

            Parameters:
                host (str): Host IP address
                port (str): Port number                
                username (str): Username for authentication/privileges
                password (str): Password for authentication
                database (str): Name of the database
                schema (str): Schema for the database
        """
        super().__init__(engine = "postgresql", host = host, port = port, database = database, username = username, password = password)
        self.engine = create_engine(self.connection_string, connect_args = {"options": f"-csearch_path={schema}"})


    @property
    def connection_string(self) -> str:
        """ Returns the connection string """
        return f"{self.engine}://{self.username}:%s@{self.host}:{self.port}/{self.database}" % urlquote(self.password)


# https://stackoverflow.com/questions/1288498/using-the-same-decorator-with-arguments-with-functions-and-methods/1288936#1288936

# Standard imports
from time import time
from typing import Callable
from datetime import datetime


def __now() -> str:
    """ Returns the current datetime as string """
    return datetime.now().strftime('%Y-%m-%d %H:%M:%S')

def __run(func: Callable, *args, **kwargs) -> tuple:
    """ Runs the given function with args and kwargs and returns the result and execution time """
    start_time = time()
    result = func(*args, **kwargs)
    return result, round(time() - start_time, 2)



class _MethodDecoratorAdaptor(object):

    def __init__(self, decorator, func):
        self.decorator = decorator
        self.func = func


    def __call__(self, *args, **kwargs):
        return self.decorator(self.func)(*args, **kwargs)


    def __get__(self, instance, owner):
        return self.decorator(self.func.__get__(instance, owner))



def auto_adapt_to_methods(decorator):
    """ Allows you to use the same decorator on methods and functions, hiding the self argument from the decorator """
    def adapt(func):
        return _MethodDecoratorAdaptor(decorator, func)
    return adapt



def status():
    """ A decorator that notifies state of the function execution and its duration """
    @auto_adapt_to_methods
    def wrapper(func: Callable):
        def wrapped(*args, **kwargs):
            try:
                print(f"[INFO - {__now()}] '{func.__name__}' started..")
                result, duration = __run(func, *args, **kwargs)
                print(f"[SUCCESS - {__now()}] '{func.__name__}' executed successfully. (Took {duration} seconds)")
                return result
            except Exception as ex:            
                print(f"[ERROR - {__now()}] '{func.__name__}' failed to execute. (details: {ex})")
        return wrapped
    return wrapper



def text(on_success: str, on_failure: str):
    """
        A decorator that shows custom text for function's execution states

        Parameters:
            on_success: Message to be shown when the function runs successfully
            on_failure: Message to be shown when the function throws and exception
    """
    @auto_adapt_to_methods
    def wrapper(func: Callable):
        def wrapped(*args, **kwargs):
            try:
                result, duration = __run(func, *args, **kwargs)
                print(f"SUCCESS - [{__now()}] {on_success}. (Took {duration} seconds)")
                return result
            except Exception as ex:            
                print(f"[ERROR - {__now()}] {on_failure}. (details: {ex})")
        return wrapped
    return wrapper



def restart_on_crash(warn_on_exception: bool = True, retries: int = None):
    """
        A decorator that restarts a function forever if it throws an exception

        Parameters:
            warn_on_exception: A boolean that indicates whether to warn the user when the function throws an exception.
            retries: The number of times to retry the function before giving up. If None, it will throw exception as normal
    """
    @auto_adapt_to_methods
    def wrapper(func: Callable):
        def wrapped(*args, **kwargs):
            _retries = 0
            while True:
                try:
                    return func(*args, **kwargs)
                except Exception as ex:
                    if retries is None: raise Exception(ex)

                    if warn_on_exception:
                        print(f"[WARNING - {__now()}] '{func.__name__}' threw an exception! restarting {_retries + 1 }/{retries} ... (details: {ex})")
                    _retries += 1   # Increment the number of retries

                # Abort the function if the number of retries is greater than the maximum number of retries
                if _retries >= retries:
                    print(f"[ERROR - {__now()}] '{func.__name__}' was retried {retries} with no lucks! aborting... (max tries: {retries})")
                    break
        return wrapped
    return wrapper



def info(message: str) -> None:
    """ Prints a message as Info """
    print(f"[INFO - {__now()}] {message}")


def success(message: str) -> None:
    """ Prints a message as Success """
    print(f"[SUCCESS - {__now()}] {message}")


def error(message: str) -> None:
    """ Prints a message as error """
    print(f"[ERROR - {__now()}] {message}")


def warning(message: str) -> None:
    """ Prints a message as warning """
    print(f"[WARNING - {__now()}] {message}")